package PObject;
# LICENSE
#
# Copyright (c) 2001 Steve Slaven. All rights reserved.
#
# This program is free software  and is provided "as is" without
# express or implied warranty. You can redistribute it and/or
# modify it under the same terms as Perl itself.
#

# This is a good base class for other objects, includes auto-binding
# for attributes and stuff.

=head1 NAME

PObject - Perl object base class

=head1 SYNOPSIS

	use PObject;
	@ISA = qw( PObject );

	%AUTOFIELDS = (
				NAME => undef,
				AGE => 35
		);

=head1 DESCRIPTION

The PObject class is meant to be inherited from.  It handles the
dirty work involved with making classes with attributes.  It handles
all attribute creation, initialization, and binding.  It is also
inheritance safe.

=head1 Initialization

Anything that inherits from the PObject will get a new() method.  This
method initializes the object and adds the field settings to the object.
When finished, it will call OBJ_NEW with the arguments to new() in
the namespace of the sub class.  Whatever is returned from OBJ_NEW
is what is returned by new().  That way, if some paramaters are not
right you can return undef and they will not get an object.

When new() is invoked, the PObject starts with it's constructor and
moves up the class tree until it reaches the class invoking it calling
the OBJ_NEW functions.  For example, if you had this tree:

	PObject
	|
	+-- Widget
		|
		+-- Field

When you called Field -> new() you'd actually be doing this:

	PObject -> OBJ_NEW()
	|
	+--> Widget -> OBJ_NEW()
		|
		+--> Field -> OBJ_NEW();

Meaning that Field will be initialized AFTER both PObject and Widget
get created.

=head1 Editable Attributes

	%AUTOFIELDS

This is the auto-field array.  Any keys in this hash become methods
and can be modified and always return the current value of the item.
For example:

	%AUTOFIELDS = (
				NAME => undef,
				AGE => 35  );

If the above is used in a class MyObj and an object is created like this:

	$o = MyObj -> new();
	$o -> Name( "Fred" );
	if( $o -> Age() == 35 )
	{
		print "Yes!!\n";
	}

The name would be set to Fred and it would print "Yes!!" because 35
is the default value for the AGE attribute.

The values are stored as $o -> { NAME } and such and can be manipulated
that way, but that should only happen inside the package since the
idea here is to hide the data types from the user.  The method names
are also case-insensitive.  So:

	$o -> Name();
	$o -> NAME();
	$o -> name();
	$o -> nAMe();

Are all the same.

=head1 Non-editable attributes

	%READFIELDS

Works the same as AUTOFIELDS but values passed to them as methods
are ignored, i.e. they B<shouldn't> be set from outside the package.

=head1 Special case for new()

	$AUTOSETTINGS

If this is true, then anything passed to new() is treated like an
initializer, i.e.:

	$o = MyObj -> new( name => 'Fred',
				age => 26 );

Is just like calling:

	$o = MyObj -> new();
	$o -> Name( 'Fred' );
	$o -> Age( 26 );

=head1 AUTHOR

Copyright (c) 2001 Steve Slaven. All rights reserved.

This program is free software  and is provided "as is" without
express or implied warranty. You can redistribute it and/or
modify it under the same terms as Perl itself.

=cut

use Carp;
use strict;
use vars qw( @ISA $VERSION %PackageDebugs );

%PackageDebugs = ();
$VERSION = "0.85";

sub new
{
    my $Check = shift;
    my $Class = ref( $Check ) || $Check;

    my %AutoSet;

    my( $HashRef, $ReadRef, $Self, @ThisISA, $UseAuto );
    my $FindPack;

    # Bail if we're a POBJ
    if( $Class eq "PObject" )
    {
        return( {} );
    }

    # This creates a ptr to the original fields
    # to check if they're OK, and inserts the defaults
    # in to this object.  Using a symbolic ref so can't have strict refs
    $FindPack = $Class;
    no strict 'refs';
    @ThisISA = eval( '@' . $FindPack . "::ISA" );
    $HashRef = eval( '\%' . $FindPack . "::AUTOFIELDS" );
    $ReadRef = eval( '\%' . $FindPack . "::READFIELDS" );
    $UseAuto = eval( '$' . $FindPack . "::AUTOSETTINGS" );
    use strict 'refs';

    if( ! $HashRef )
    {
	$HashRef = {};
    }

    if( ! $ReadRef )
    {
	$ReadRef = {};
    }

    # First check through the ISA ref to find any parents.  Call new on them.
    if( @ThisISA )
    {
        $Self = ( shift( @ThisISA ) -> new() );
    }
    else
    {
        $Self = {};
    }

    if( ! defined( $PackageDebugs{ $Class } ) )
    {
	$PackageDebugs{ $Class } = 0;
    }
    bless $Self, $Class;

    # Add in the things
    $Self -> { _POBJ_DEBUG } = 0;

    for ( keys( %{ $HashRef } ) )
    {
	#$Self -> _dmesg( "Adding autofield: $_", 3 );
        $Self -> { $_ } = $HashRef -> { $_ };
	$Self -> { _POBJ_AUTOFIELD } -> { $_ } = 1;
    }

    for ( keys( %{ $ReadRef } ) )
    {
	#$Self -> _dmesg( "Adding readfield: $_", 3 );
        $Self -> { $_ } = $ReadRef -> { $_ };
	$Self -> { _POBJ_READFIELD } -> { $_ } = 1;
    }
   
    # Do the auto-settings
    if( $UseAuto )
    {
        $Self = $Self -> OBJ_NEW();

	if( $Self )
	{
		#$Self -> _dmesg( "Using AUTOSETTINGS", 3 );
		%AutoSet = @_;
        	for ( keys( %AutoSet ) )
        	{
	    		#$Self -> _dmesg( $_ . ":" . $AutoSet{ $_ }, 3 );
            		$Self -> $_( $AutoSet{ $_ } );
        	}
	}
    }
    else
    {
        $Self = $Self -> OBJ_NEW( @_ );
    }

    return( $Self );
}

sub AUTOLOAD
{
    my $Self = shift;
    my $Arg = shift;
    
    my $Field = $PObject::AUTOLOAD;
    
    $Field =~ s/.*://;
    my $HookMethod;
    
    if( ! ref( $Self ) )
    {
	confess( "Called from package name: " . $Self . " -> " . $Field );
    }

    #$Self -> _dmesg( "AUTOLOAD is: '" . $Field . "'", 9 );
    #$Self -> _dmesg( "Arg ref is: '" . ref( $Arg ) . "'", 9 );

    $Field = uc( $Field );

    # Check for HOOK
    if( $Field =~ /^OBJ_HOOK/ )
    {
	#$Self -> _dmesg( "Failed hook, returning value", 9 );
	return( $Arg );
    }
    
    #$Self -> _dmesg( "Checking for AUTOFIELD: " . $Field, 9 );
    if( $Self -> { _POBJ_AUTOFIELD } -> { $Field } )
    {
	# OK do it
	#$Self -> _dmesg( "Found AUTOFIELD: " . $Field . "='" . $Self -> { $Field } . "'", 9 );

	if( defined( $Arg ) )
	{
	    $HookMethod = "OBJ_HOOK_" . $Field;
	    $Arg = $Self -> $HookMethod( $Arg );
	
	    if( defined( $Arg ) )
	    {
		$Self -> { $Field } = $Arg;
		#$Self -> _dmesg( "Set value: '" . $Self -> { $Field } . "'", 9 );
	    }
	}
	
	return( $Self -> { $Field } );
    }
    else
    {
	# Check for read-only field
	if( $Self -> { _POBJ_READFIELD } -> { $Field } )
	{
	    return( $Self -> { $Field } );
	}
	else
	{
	    #$Self -> _dmesg( "Trying OBJ_AUTOLOAD", 9 );
	    return( $Self -> OBJ_AUTOLOAD( $Field, @_ ) );
	}
    }
}

sub DESTROY
{
    # This needs to be defined because of the AUTOLOAD thing.
}

sub OBJ_AUTOLOAD
{
    # Bad call freak.
    my $Self = shift;
    confess( "Uneditable or unknown attribute called: " . shift() );
}

sub OBJ_NEW
{
    return( shift() );
}


# Debugging functions
sub _debug
{
    # Sets the debugging level
    my $Self = shift;
    my $Debug = shift;

    my $ClassName = ref( $Self ) || $Self;

    if( ref( $Self ) )
    {
	$Self -> { _POBJ_DEBUG } = $Debug;
	#$Self -> _dmesg( "DEBUG level set for object", 1 );
    }
    else
    {
	$PackageDebugs{ $ClassName } = $Debug;
	#$Self -> _dmesg( "DEBUG level set for package", 1 );
    }
	
    $Self -> _dmesg( "Debug level set to $Debug", 1 );
}

sub _dmesg
{
    my $Self = shift;
    my $Message = shift;
    my $Level = shift;

    my $ClassName = ref( $Self ) || $Self;
    my $Build = ( ref( $Self ) || $Self ) . "[" . $Level . "]) " . $Message;

    if( $PackageDebugs{ $ClassName } >= $Level )
    {
	# Spew regardless
	carp( $Build );
    }
    else
    {
	# Check object level
	if( ref( $Self ) && ( $Self -> { _POBJ_DEBUG } >= $Level ) )
	{
	    carp( $Build );
	}
    }
}

sub _dfatal
{
    # This is just a croak type call
    my $Self = shift;
    my $Message = shift;
    my $Perspective = shift;

    $Self -> _dmesg( $Message, -1, $Perspective );
    croak( "Fatal error..." );
}

sub ISA
{
    my $Self = shift;
    
    return( ref( $Self ) )
}

1;
